#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_sf_gamma.h>
#include "type.h"
#include "table.h"

#define EULER   0.57721566490153286060651209008240243104215933593992
#define GAMMA1 -0.07281584548367672486058637587490131913773633833433
#define INIT_DONE 49503343

int state = 0;
MYFLOAT *table;
MYFLOAT *C;
unsigned long tablesize;

/*
 * initialize the table of the fixed values
 * PARAMS:
 * - dim: the needed dimension
 * - verbose: debugging parameter
 * RETURNS
 * - the availble dimension of the table
 */
unsigned long init(unsigned long dim, int verbose) {
  
  // timing vars
  clock_t begin;
  clock_t end;
  double time_spent;

  if (state == INIT_DONE && dim <= tablesize) {
      return dim;
  }

  if (state == INIT_DONE) {
    /* reallocate table and computing array */
    free((void *) table);
    free((void *) C);
  }
  
  begin = clock();

  if ((table = malloc(dim * sizeof(MYFLOAT))) == NULL) {
    fprintf(stderr, "Unable to allocate table! Giving up\n");
    exit(1);
  }
  
  setTableDim(dim);
  tablesize = loadTable(table, dim);
  if (tablesize < dim) {
    fprintf(stderr, "WARNING: read only %lu elements from table.txt (expected=%lu)\n",
	    tablesize, dim);
    if (tablesize % 2 == 1) {
      tablesize--;
    }
  }
  
  state = INIT_DONE;
  if (verbose)
    printf("init() done\n");

  if ((C = (MYFLOAT *) calloc( (tablesize + 1), sizeof(MYFLOAT))) == NULL) {
    fprintf(stderr, "Unable to allocate C (sum)! Giving up\n");
    exit(1);
  }

  fprintf(stdout, "INFO: precomputed table allocated %1lu elements\n", tablesize);

  end = clock();
  time_spent = (double)(end - begin) / CLOCKS_PER_SEC;

  fprintf(stdout, "INFO: init() time (seconds) = %lf\n", time_spent);
  
  return tablesize;
}

/* 
 * Elevamento a potenza in aritmetica modulare
 * PARAMS:
 * - base: the number to be elevated
 * - exp: the exponent to elevate
 * - mod: the module
 */
long int lpow_mod(long int base, long int exp, long int mod)
{
    long int result = 1;
    base = base % mod;
    for (;;)
    {
      if (exp & 1) {
	result *= base;
	result %= mod;
      }
      exp >>= 1;
      if (!exp)
	break;
      base *= base;
      base = base % mod;
    }

    return result;
}

/*
 * debug e confronti
 */
long int pmod(long int base, long int exp, long int mod) {
  long int result = 1;
  for(int i = 1; i <= exp; i++) {
    result = (result * base) % mod;
  }
  return result;
}

/* 
 * somma con algoritmo di Kahan (https://en.wikipedia.org/wiki/Kahan_summation_algorithm)
 * PARAMS:
 * - A: pointer to array to sum 
 * - n: number of elements of the array (0...n-1)
 * RETURNS:
 * - the sum of the elements of the array A
 * FEATURES:
 * - leave A untouched
 */
MYFLOAT sumKahan(MYFLOAT *A, unsigned long n) {
  MYFLOAT sum = 0.0;
  MYFLOAT c = 0.0;
  MYFLOAT y = 0.0;
  MYFLOAT t = 0.0;
  
  for (long i = 0; i < n; i++) {
    y = A[i] - c;
    t = sum + y;
    c = (t - sum) - y;
    sum = t;
  }

  return sum;
}

/* 
 * somma con algoritmo 'pairwise' 
 * (https://en.wikipedia.org/wiki/Pairwise_summation)
 * e usando l'algoritmo di Kahan per il blocco minimo
 * PARAMS:
 * - A: pointer to array to sum 
 * - n: number of elements of the array (0...n-1)
 * RETURNS:
 * - the sum of the elements of the array A
 * FEATURES:
 * - leave A untouched
 */
MYFLOAT sumPairwiseKahan(MYFLOAT *A, unsigned long n) {
  unsigned long N = 32;
  unsigned long m = 0;
  MYFLOAT sum = 0.0;
  
  if (n < N) {
    sum = 0.0;
    for (unsigned long i = 0; i < n; i++) {
      sum += A[i];
    }
    // return sumKahan(A, n);
  } else {
    
    /*
      n = 5: m=2 divido in A[0,1], A+2[0,1,2]
    */
    
    m = n / 2;
    
    sum = sumPairwiseKahan(A,m) + sumPairwiseKahan(A+m, n - m);
  }

  return sum;
}


/* 
 *  effettua la somma di un array a con n elementi dicotomicamente
 *  e 'in place' quindi l'array viene modificato. Nel primo elemento la somma.
 *  restituisce la somma. 
 */
MYFLOAT sumDicotomic(MYFLOAT *A, unsigned long n) {
  long limit;
  long writer, reader;
  
  reader = 0;
  writer = 0;
  limit  = n - 1; /* limit = indice dell'ultimo elemento di A */
  while (limit > 0) {
    while (reader + 1 <= limit) {
	A[writer++] = A[reader] + A[reader + 1];
	reader+=2;
    }

    if (reader <= limit) {
      A[writer++] = A[reader++];
    }

    /* printf("limit = %lu, reader = %lu, writer = %lu\n", limit, reader, writer); */
    
    limit = writer - 1;
    writer = 0;
    reader = 0;
  }
  return A[0];
}

/*
 * Calcolo della funzione S(x)
 * PARAMS:
 * - x: ascissa
 * - precision: the desired precision (bits)
 * - verbose: debug parameter
 * RETURNS:
 * the value of y = S(x)
 */
MYFLOAT s(MYFLOAT aq, int precision, int verbose) {

  MYFLOAT  oneminusaq = 1.0 - aq;
  MYFLOAT  logaq = LOG(aq);
  MYFLOAT  logoneminusaq = LOG(oneminusaq);
  // MYFLOAT  logsin = LOG(SIN(PIGRECO * aq));
  // MYFLOAT  logpi = LOGPI;
  // MYFLOAT  difflog = logpi - logsin;
  MYFLOAT  coeff = (precision + 1) * LOG2;
  MYFLOAT  r1;
  MYFLOAT  square;
  MYFLOAT  correctionvalue;
  MYFLOAT  u;
  MYFLOAT  G;
  unsigned long int rp;
  
  if (aq > 0.5) { // a/q >1/2, (1-a/q) <1/2 // equations 36-37 of the paper
    r1 = CEIL((coeff + ABS(logaq)) / ABS(logoneminusaq)); // number of summands
    square = oneminusaq * oneminusaq;
    correctionvalue = logoneminusaq * logoneminusaq;
  } else {
    r1 = CEIL((coeff + ABS(logoneminusaq)) / ABS(logaq));
    square = aq * aq;
    correctionvalue = logaq * logaq;
  }  

  rp = max(precision + 10, r1);
  
  rp = init((unsigned long int) rp, verbose);
  
  C[0] = 0.0;
  C[1] = 0.0;

  u = square;
  for(long int k = 2; k <= rp; k+=2) {
    C[k] = u * table[k - 2];
    // fprintf(stderr, "S[%1ld] = %32s, ", k, fstr(C[k]));
    // fprintf(stderr, "v[%1ld] = %32s\n", k - 2, fstr(table[k - 2]));
    u *= square;
  }

  // fprintf(stderr, "%32s\n", fstr(aq));

  G = sumKahan(C, rp);

  // fprintf(stderr, "%32s\n", fstr(G));

  G = 2 * G;
  G = - (2 * G + correctionvalue);

  // fprintf(stderr, "%32s\n", fstr(G));
  // fprintf(stderr, "%lu\n", rp);
  // fprintf(stderr, "%32s\n", fstr(square));
  // fprintf(stderr, "%32s\n", fstr(correctionvalue));
  // fprintf(stderr, "-----\n"); 
  return G;
  
}
